/*
								+---------------------------------+
								|                                 |
								|        ***   Main   ***         |
								|                                 |
								|  Copyright   -tHE SWINe- 2002  |
								|                                 |
								|            Main.cpp             |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include <malloc.h>
#include <stdio.h>
#include <math.h>

#include "Directx.h"
#include "Matrix.h"
#include "Font_eng.h"
#include "Load_3ds.h"
#include "Raster.h"
#include "Main.h"


#define APPNAME "ZBuffer_Engine"
#define WNDNAME "Z-Buffer Engine 2.0"
static int Width = 512;
static int Height = 384;

unsigned __int32 *video;
unsigned __int32 *buffer;

#define PI 3.1415926

static World *svet;
static Matrix cmatrix;
static Texture *obrazek[2];

int WINAPI WinMain(HINSTANCE hInstance,
	HINSTANCE hPrevInstance, PSTR szCmdline, int nCmdShow)
{
	WNDCLASSEX wndclass;
	char szAppName[] = APPNAME;
	HWND hwnd;
	MSG msg;

	wndclass.cbSize			= sizeof(wndclass);
	wndclass.style			= CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc	= Wokno;
	wndclass.cbClsExtra		= 0;
	wndclass.cbWndExtra		= 0;
	wndclass.hInstance		= hInstance;
	wndclass.hIcon			= LoadIcon(hInstance, IDI_APPLICATION);
	wndclass.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground	= (HBRUSH)GetStockObject(BLACK_BRUSH);
	wndclass.lpszMenuName	= 0;
	wndclass.lpszClassName	= szAppName;
	wndclass.hIconSm		= LoadIcon(NULL, IDI_APPLICATION);
	RegisterClassEx(&wndclass);
	hwnd = CreateWindow(szAppName,
		WNDNAME, WS_DLGFRAME,
		CW_USEDEFAULT, CW_USEDEFAULT,
		Width, Height,
		NULL, NULL,	hInstance, NULL);
	ShowWindow(hwnd, nCmdShow);
	UpdateWindow(hwnd);
	// vytvo okno

	ShowCursor(FALSE);
	// skryje kurzor

	Matrix_Init(&cmatrix);
	Matrix_Rotate_Z(&cmatrix, (float)-PI);
	Matrix_Rotate_X(&cmatrix, (float)-PI / 6);
	Matrix_Translate(&cmatrix, 0, 0, -200);
	// kamera ok

	static int i;
	if((svet = Load_3DS("Z_Test.3ds")) == NULL)
		return 0;
	for(i = 0; i < svet->n_object_num; i ++)
		Matrix_Init(&svet->object[i].m_object);
	// svet ok

	Init_Engine(Width, Height);
	Init_Font_Engine(Width, Height);
	// piprav engine

	if((obrazek[0] = Load_Texture("Tex.bmp")) == NULL)
		return 0;
	if((obrazek[1] = Load_Texture("Tex2.bmp")) == NULL)
		return 0;
	// nahraje textury

	if((buffer = (unsigned __int32*) malloc(Width * Height * 4)) == NULL)
		return 0;
	// naalokuje pam pro obraz

	SetTimer(hwnd, 0, 20, NULL);
	// nastav timer

	InitDirectX(hwnd, Width, Height);
	// inicializuje directX

	while(GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	TerminateDirectX();
	free((void*)buffer);
	Terminate_Engine();
	Terminate_Font_Engine();
	// uvoln pam

	ShowCursor(TRUE);
	// obnov kurzor

	return msg.wParam;
}

void Quincunx(unsigned __int32 *src, unsigned __int32 *dest)
{
	int w2, h2, w1;
	int x, y;

	w2 = Width - 1;
	h2 = Height - 1;
	w1 = w2 + 1;

	y = Width * Height;
	for(x = 0; x < y; x ++)
		src[x] = ((src[x] >> 2) & 0x3f3f3f);

	for(y = 0; y < h2; y ++, dest += Width, src += Width) {
		for(x = 0; x < w2; x ++)
			dest[x] = src[x] + src[x + 1] + src[x + w1] + src[x + w2];
		dest[x] = src[x];
	}
	memcpy(dest, src, Width << 2);

	// obrazovy filtr ATI ? Quincunx
}


LRESULT CALLBACK Wokno (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	float speeds[] = {2, 1, 0.5};
	static float a = 0;
	static int j = 0;
	char str[256];
	int i;

	switch (iMsg) {
	case WM_KEYDOWN :
		if(wParam == VK_ESCAPE) {
			SendMessage(hwnd, WM_CLOSE, 0, 0);
			return 0;
		}
		return 0;

	case WM_TIMER :
		memset(buffer, 0, 4 * Width * Height);
		// vycisti obrazovku

		a += (float)0.1;
		if(a > 6.28)
			a = 0;
		// pot hel

		Clean_ZBuffer();
		// pred kazdym snimkem se musi provest

		for(i = 0; i < svet->n_object_num; i ++) {
			Matrix_Rotate_X(&svet->object[i].m_object, (float)PI / 100 * speeds[i & 3]);
			Matrix_Rotate_Y(&svet->object[i].m_object, (float)(PI / 100 * sin(a)) * speeds[i & 3]);
			DrawObject(&svet->object[i].m_object, &cmatrix,
				&svet->object[i], buffer, obrazek[i & 3]);
		}
		// kresli, rotuje objekty (1. se rotuje 2 rychleji...)
		// (i & 3) = (i % 4)

		if(InitVideoBuffer(&video)) {
			///
			//Quincunx(video, buffer);
			memcpy(video, buffer, 4 * Width * Height);
			/// zkopruje obraz

			sprintf(str, "%d.th Frame", j ++);
			String_Out(str, 10, 10, video);
			// vypise cislo snimku
		
			FlipVideoBuffer();
		}

		return 0;

	case WM_DESTROY:
		PostQuitMessage (0);
		return 0;
	}
	return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}





